var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
export function FormControlMixin(SuperClass) {
    var _FormControl_instances, _FormControl_focused, _FormControl_forceError, _FormControl_touched, _FormControl_isCheckedElement_get, _FormControl_initFormControl, _FormControl_commitValue, _FormControl_shouldShowError, _FormControl_onFocus, _FormControl_onBlur, _FormControl_onInvalid, _FormControl_setValue, _FormControl_validate;
    class FormControl extends SuperClass {
        /* eslint-disable  @typescript-eslint/no-explicit-any */
        constructor(...args) {
            super(...args);
            _FormControl_instances.add(this);
            /** The ElementInternals instance for the control. */
            this.internals = this.attachInternals();
            /**
             * Keep track of if the control has focus
             * @private
             */
            _FormControl_focused.set(this, false);
            /**
             * Exists to control when an error should be displayed
             * @private
             */
            _FormControl_forceError.set(this, false);
            /**
             * Toggles to true whenever the element has been focused. This property
             * will reset whenever the control's formResetCallback is called.
             * @private
             */
            _FormControl_touched.set(this, false);
            /**
             * The controls' form value. As this property is updated, the form value
             * will be updated. If a given control has a `checked` property, the value
             * will only be set if `checked` is truthy.
             */
            this.value = '';
            /**
             * Set this[touched] and this[focused]
             * to true when the element is focused
             * @private
             */
            _FormControl_onFocus.set(this, () => {
                __classPrivateFieldSet(this, _FormControl_touched, true, "f");
                __classPrivateFieldSet(this, _FormControl_focused, true, "f");
                __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
            });
            /**
             * Reset this[focused] on blur
             * @private
             */
            _FormControl_onBlur.set(this, () => {
                __classPrivateFieldSet(this, _FormControl_focused, false, "f");
                /**
                 * Set forceError to ensure error messages persist until
                 * the value is changed.
                 */
                if (!this.validity.valid && __classPrivateFieldGet(this, _FormControl_touched, "f")) {
                    __classPrivateFieldSet(this, _FormControl_forceError, true, "f");
                }
                const showError = __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
                if (this.validationMessageCallback) {
                    this.validationMessageCallback(showError ? this.validationMessage : '');
                }
            });
            /**
             * For the show error state on invalid
             * @private
             */
            _FormControl_onInvalid.set(this, () => {
                __classPrivateFieldSet(this, _FormControl_forceError, true, "f");
                __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
            });
            this.addEventListener('focus', __classPrivateFieldGet(this, _FormControl_onFocus, "f"));
            this.addEventListener('blur', __classPrivateFieldGet(this, _FormControl_onBlur, "f"));
            this.addEventListener('invalid', __classPrivateFieldGet(this, _FormControl_onInvalid, "f"));
        }
        /** Wires up control instances to be form associated */
        static get formAssociated() {
            return true;
        }
        static get validators() {
            return this.formControlValidators || [];
        }
        /**
         * Allows the FormControl instance to respond to Validator attributes.
         * For instance, if a given Validator has a `required` attribute, that
         * validator will be evaluated whenever the host's required attribute
         * is updated.
         */
        static get observedAttributes() {
            const validatorAttributes = this.validators.map((validator) => validator.attribute);
            const observedAttributes = super.observedAttributes || [];
            /** Make sure there are no duplicates inside the attributes list */
            const attributeSet = new Set([...observedAttributes, ...validatorAttributes]);
            return [...attributeSet];
        }
        /**
         * Return the validator associated with a given attribute. If no
         * Validator is associated with the attribute, it will return null.
         */
        static getValidator(attribute) {
            return this.validators.find((validator) => validator.attribute === attribute) || null;
        }
        /** Return a reference to the control's form */
        get form() {
            return this.internals.form;
        }
        /**
         * Will return true if it is recommended that the control shows an internal
         * error. If using this property, it is wise to listen for 'invalid' events
         * on the element host and call preventDefault on the event. Doing this will
         * prevent browsers from showing a validation popup.
         */
        get showError() {
            return __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
        }
        /**
         * Forward the internals checkValidity method
         * will return the valid state of the control.
         */
        checkValidity() {
            return this.internals.checkValidity();
        }
        /**
         * The validation message shown by a given Validator object. If the control
         * is in a valid state this should be falsy.
         */
        get validationMessage() {
            return this.internals.validationMessage;
        }
        /**
         * The element's validity state after evaluating the control's Validators.
         * This property implements the same patterns as the built-in constraint
         * validation strategy.
         */
        get validity() {
            return this.internals.validity;
        }
        attributeChangedCallback(name, oldValue, newValue) {
            var _a;
            (_a = super.attributeChangedCallback) === null || _a === void 0 ? void 0 : _a.call(this, name, oldValue, newValue);
            /**
             * Check to see if a Validator is associated with the changed attribute.
             * If one exists, call control's validate function which will perform
             * control validation.
             */
            const proto = this.constructor;
            const validator = proto.getValidator(name);
            if (validator && this.validationTarget) {
                __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_validate).call(this, this.value);
            }
        }
        connectedCallback() {
            var _a;
            (_a = super.connectedCallback) === null || _a === void 0 ? void 0 : _a.call(this);
            /** Initialize the form control and perform initial validation */
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_initFormControl).call(this);
        }
        disconnectedCallback() {
            var _a;
            (_a = super.disconnectedCallback) === null || _a === void 0 ? void 0 : _a.call(this);
            /**
             * Remove the event listeners that toggles the touched and focused states
             */
            this.removeEventListener('focus', __classPrivateFieldGet(this, _FormControl_onFocus, "f"));
            this.removeEventListener('blur', __classPrivateFieldGet(this, _FormControl_onBlur, "f"));
            this.removeEventListener('invalid', __classPrivateFieldGet(this, _FormControl_onInvalid, "f"));
        }
        /** Reset control state when the form is reset */
        formResetCallback() {
            this.resetFormControl();
        }
        /**
         * Resets a form control to its initial state
         */
        resetFormControl() {
            if (__classPrivateFieldGet(this, _FormControl_instances, "a", _FormControl_isCheckedElement_get)) {
                this.checked = false;
            }
            else {
                this.value = '';
            }
            __classPrivateFieldSet(this, _FormControl_touched, false, "f");
            __classPrivateFieldSet(this, _FormControl_forceError, false, "f");
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
        }
    }
    _FormControl_focused = new WeakMap(), _FormControl_forceError = new WeakMap(), _FormControl_touched = new WeakMap(), _FormControl_onFocus = new WeakMap(), _FormControl_onBlur = new WeakMap(), _FormControl_onInvalid = new WeakMap(), _FormControl_instances = new WeakSet(), _FormControl_isCheckedElement_get = function _FormControl_isCheckedElement_get() {
        return this.hasOwnProperty('checked') || this.constructor.prototype.hasOwnProperty('checked');
    }, _FormControl_initFormControl = function _FormControl_initFormControl() {
        /** Closed over variable to track value changes */
        let value = this.value || '';
        /** Look to see if '`checked'` is on the control's prototype */
        const hasChecked = __classPrivateFieldGet(this, _FormControl_instances, "a", _FormControl_isCheckedElement_get);
        /**
         * The FormControlMixin writes the value property on the element host
         * this checks to see if some other object in the prototype chain
         * has a getter/setter for value and saves a reference to those.
         *
         * We do this to make sure that we don't overwrite behavior of an object
         * higher in the chain.
         */
        let descriptor = Object.getOwnPropertyDescriptor(this, 'value');
        /**
         * Many libraries like Lit will write properties to the element's
         * prototype. This will pull the get/set from the prototype descriptor
         * if it is available.
         */
        if (Object.getOwnPropertyDescriptor(this.constructor.prototype, 'value')) {
            descriptor = Object.getOwnPropertyDescriptor(this.constructor.prototype, 'value');
        }
        /** Make sure to defer to the parent */
        const set = descriptor && descriptor.set;
        const get = descriptor && descriptor.get;
        /** Define the FormControl's value property */
        Object.defineProperty(this, 'value', {
            get() {
                /** If a getter already exists, make sure to call it */
                if (get) {
                    return get.call(this);
                }
                return value;
            },
            set(newValue) {
                /** Save a reference to the new value to use later if necessary */
                value = newValue;
                __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_commitValue).call(this, newValue);
                /** If a setter already exists, make sure to call it */
                if (set) {
                    set.call(this, newValue);
                }
            }
        });
        /**
         * If checked already exists on a prototype, we need to monitor
         * for changes to that property to ensure the proper value is set on the
         * control's form.
         *
         * TODO: Justin Fagnani cautioned that this might not scale well. Maybe
         * this should be a direct check against the value of checked ...
         */
        if (hasChecked) {
            /**
             * As with value, save a reference to the getter/setter if they already
             * exist in the prototype chain
             */
            let descriptor = Object.getOwnPropertyDescriptor(this, 'checked');
            if (this.constructor.prototype.hasOwnProperty('checked')) {
                descriptor = Object.getOwnPropertyDescriptor(this.constructor.prototype, 'checked');
            }
            const get = descriptor && descriptor.get;
            const set = descriptor && descriptor.set;
            /** Close over the initial value to use in the new getter/setter */
            let { checked } = this;
            Object.defineProperty(this, 'checked', {
                get() {
                    /** If a getter exists, use it */
                    if (get) {
                        return get.call(this);
                    }
                    return checked;
                },
                set(newChecked) {
                    /** Updated closure value */
                    checked = newChecked;
                    if (newChecked) {
                        /** If truthy, set the form value to the instance's value */
                        __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_setValue).call(this, this.value);
                    }
                    else {
                        /** If falsy, remove the instance's form value */
                        __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_setValue).call(this, null);
                    }
                    /** If a setter exists, use it */
                    if (set) {
                        set.call(this, newChecked);
                    }
                }
            });
        }
        __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_commitValue).call(this, value);
    }, _FormControl_commitValue = function _FormControl_commitValue(value) {
        if (__classPrivateFieldGet(this, _FormControl_instances, "a", _FormControl_isCheckedElement_get) && this.checked) {
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_setValue).call(this, value);
        }
        else if (__classPrivateFieldGet(this, _FormControl_instances, "a", _FormControl_isCheckedElement_get) && !this.checked) {
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_setValue).call(this, null);
        }
        else {
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_setValue).call(this, value);
        }
    }, _FormControl_shouldShowError = function _FormControl_shouldShowError() {
        if (this.hasAttribute('disabled')) {
            return false;
        }
        const showError = __classPrivateFieldGet(this, _FormControl_forceError, "f") || (__classPrivateFieldGet(this, _FormControl_touched, "f") && !this.validity.valid && !__classPrivateFieldGet(this, _FormControl_focused, "f"));
        if (showError) {
            this.internals.states.add('--show-error');
        }
        else {
            this.internals.states.delete('--show-error');
        }
        return showError;
    }, _FormControl_setValue = function _FormControl_setValue(value) {
        __classPrivateFieldSet(this, _FormControl_forceError, false, "f");
        this.internals.setFormValue(value); /** Typed to string to work with polyfill */
        if (this.valueChangedCallback) {
            this.valueChangedCallback(value);
        }
        __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_validate).call(this, value);
        const showError = __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
        if (this.validationMessageCallback) {
            this.validationMessageCallback(showError ? this.validationMessage : '');
        }
    }, _FormControl_validate = function _FormControl_validate(value) {
        const proto = this.constructor;
        const validity = {};
        let validationMessage = '';
        let isValid = true;
        proto.validators.forEach((validator) => {
            /** Get data off the Validator */
            const { message, callback } = validator;
            /** If a key is not set, use `customError` as a catch-all */
            const key = validator.key || 'customError';
            /** Invoke the Validator callback with the instance and the value */
            const valid = callback(this, value);
            /**
             * Invert the validity because we are setting the new property
             * on the new ValidityState object
             */
            validity[key] = !valid;
            if (valid === false) {
                isValid = false;
                let messageResult = '';
                /**
                 * The Validator interfaces allows for the message property
                 * to be either a string or a function. If it is a function,
                 * we want to get the returned value to use when calling
                 * ElementInternals.prototype.setValidity.
                 *
                 * If the Validator.message is a string, use it directly. However,
                 * if a control has a ValidityCallback, it can override the error
                 * message for a given validity key.
                 */
                if (this.validityCallback && this.validityCallback(key)) {
                    messageResult = this.validityCallback(key);
                }
                else if (message instanceof Function) {
                    messageResult = message(this, value);
                }
                else if (typeof message === 'string') {
                    messageResult = message;
                }
                validationMessage = messageResult;
            }
        });
        /**
         * In some cases, the validationTarget might not be rendered
         * at this point, if the validationTarget does exist, proceed
         * with a call to internals.setValidity. If the validationTarget
         * is still not set, we essentially wait a tick until it is there.
         *
         * If the validityTarget does not exist even after the setTimeout,
         * this will throw.
         */
        if (isValid) {
            this.internals.setValidity({});
        }
        else if (this.validationTarget) {
            this.internals.setValidity(validity, validationMessage, this.validationTarget);
        }
        else {
            /**
             * If the validationTarget is not set, the user can decide how they would
             * prefer to handle focus when the field is validated.
             *
             * TODO: Document this edge case
             */
            this.internals.setValidity(validity, validationMessage);
            /**
             * It could be that a give component hasn't rendered by the time it is first
             * validated. If it hasn't been, wait a bit and add the validationTarget
             * to the setValidity call.
             *
             * TODO: Document the edge case that an element doesn't have a validationTarget
             * and must be focusable some other way
             */
            let tick = 0;
            const id = setInterval(() => {
                if (tick >= 100) {
                    clearInterval(id);
                }
                else if (this.validity.valid) {
                    clearInterval(id);
                }
                else if (this.validationTarget) {
                    this.internals.setValidity(this.validity, this.validationMessage, this.validationTarget);
                    clearInterval(id);
                }
                tick += 1;
            }, 0);
        }
    };
    return FormControl;
}
//# sourceMappingURL=FormControlMixin.js.map